#include "dill.h"

index_block * DILL::alloc_block(index_block * parent){
	index_block *	block = NULL;
	
	block = (index_block *) getmem(sizeof(index_block));
	mem_used += sizeof(index_block);
	
	if(block){
		last_block = block;
		memset(block, 0, sizeof(index_block));
		block->offset = num_blocks * NUM_INDEXES;
		num_blocks++;
		block->num_used = 0;
		
		if(parent){
			parent->child = block;
			block->parent = parent;
		}
	}
	
	return block;
}

uint32 DILL::alloc_data(index_block * block, void * data, uint32 size){
	void *			tmp = NULL;
	uint32			i = 0;
	
	tmp = getmem(size);
	mem_used += size;
	
	if(tmp){
		memcpy(tmp, data, size);
		
		while(block->indexes[i]) i++;
		
		block->indexes[i] = tmp;
		block->num_used++;
		return (i + block->offset);
	}
	
	return ALLOC_BAD;
}

DILL::DILL(){
	num_blocks = 0;
	mem_used = 0;
	first_block = alloc_block(NULL);
	usr_free = NULL;
}

DILL::~DILL(){
	index_block *	block = first_block;
	void *			ptr;
	uint32			i;
	
//	printf("Deallocating memory... ");
	
	while(block){
		
		for(i = 0;i < NUM_INDEXES;i++){
			ptr = block->indexes[i];
			if(ptr){
				if(usr_free) mem_used -= usr_free(ptr);
				freemem(ptr);
			}
		}
		
		ptr = block;
		block = (index_block *) block->child;
		freemem(ptr);
	}

//	printf("done\n");
}

uint32 DILL::push(void * data, uint32 size){
	index_block *	block = first_block;

	while(block){
		if(block->num_used < NUM_INDEXES) break;
		block = (index_block *) block->child;
	}
	
	if(block){
		return alloc_data(block, data, size);
	}else{
		if(alloc_block(last_block)){
			return alloc_data(last_block, data, size);
		}
	}

	return ALLOC_BAD;
}

void * DILL::pop(uint32 index){
	index_block *	block = first_block;
	
	if(index < num_blocks * NUM_INDEXES){
		
		while(block->offset + NUM_INDEXES <= index){
			block = (index_block *) block->child;
		}
		
		return block->indexes[index - block->offset];
	}

	return NULL;
}

void DILL::erase(uint32 index){
	index_block *	block = first_block;
	void *			ptr;
	uint32			i;
	
	if(index < num_blocks * NUM_INDEXES){
		
		while(block->offset + NUM_INDEXES <= index){
			block = (index_block *) block->child;
		}
		
		i = index - block->offset;
		ptr = block->indexes[i];
		
		if(ptr){
			if(usr_free) mem_used -= usr_free(ptr);
			freemem(ptr);
			block->indexes[i] = NULL;
			block->num_used--;
		}
	}
}

void DILL::show(){
	uint32	i = 0, j;
	index_block *	block = first_block;
	printf("\nAllocated blocks: %d Memory used: %d bytes\n\n\n", num_blocks, mem_used);
	
	while(i < num_blocks){
		printf("Block\t%04d\t%p:\n\tOffset\t%08d\n\tUsed\t%02d\n\tParent\t%p\n\tChild\t%p\n\tIndexes:", i, block, block->offset, block->num_used, block->parent, block->child);
		for(j = 0;j < NUM_INDEXES;j++){
			if(j % 5 == 0) printf("\n\t");
			printf("%02d %p ", j, block->indexes[j]);
		}
		putchar('\n');
		block = (index_block *) block->child;
		i++;
	}
}

void DILL::set_usr_free(uint32 (* func)(void *)){
	usr_free = func;
}

bool DILL::fill(void * data, uint32 count, uint32 size){
	uint32	i;

	for(i = 0;i < count;i++){
		if(push(data, size) == ALLOC_BAD) return false;
	}
	
	return true;
}

void DILL::get_used(uint32 * blocks, uint32 * mem){
	*blocks = num_blocks;
	*mem = mem_used;
}
